"""The vision pipeline abstraction helps chain image processing operations as sequence of steps. Each step consumes and produces a `FrameSet` which typically contains a source image and derivative metadata and images."""



"""This module contains common streamlit utilities that are used across the example Apps in this repo.
It's only intended for the example streamlit Apps. When using the SDK in your own project, you don't need to use this module.
"""
import logging
import os
from typing import Any, Optional, Tuple

_DEFAULT_API_KEY_ENV_VAR = "LANDINGAI_API_KEY"

_LOGGER = logging.getLogger(__name__)


def _import_st() -> Any:
    """Import streamlit and raise an error if it fails."""
    try:
        import streamlit as st

        return st
    except ImportError as e:
        raise ValueError(
            """Failed to import streamlit due to missing the `streamlit` dependency.
This is likely because you are trying to use a function in this module outside of the example Apps in this repo. But this function is only intended for example Apps of this repo.
If you are using this function in other environment, please install streamlit manually first, e.g. 'pip install streamlit'.
If you are running one of the example Apps of this repo, please follow the installation instructions of that App.
    """
        ) from e


def setup_page(page_title: str) -> None:
    """Common setup code for streamlit pages.
    This function should be called only once at the beginning of the page.
    """
    level = os.environ.get("LOGLEVEL", "INFO").upper()
    logging.basicConfig(
        level=level,
        format="%(asctime)s %(filename)s %(funcName)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    st = _import_st()
    st.set_page_config(
        page_title=page_title,
        page_icon="./examples/apps/assets/favicon.ico",
        layout="centered",
        initial_sidebar_state="auto",
    )
    # Hide the default menu button (on the top right corner) and the footer
    hide_footer_style = """
        <style>
        .reportview-container .main footer {visibility: hidden;}
        #MainMenu {
                visibility: hidden;
            }

        footer {
                visibility: hidden;
            }
        </style>
        """
    st.markdown(hide_footer_style, unsafe_allow_html=True)


def render_svg(svg: str, margin_bottom: int = 3) -> None:
    """Renders the given svg string.
    This is a workaround as st.image doesn't seem to work.
    See: https://github.com/streamlit/streamlit/issues/275
    """
    st = _import_st()
    import base64

    b64 = base64.b64encode(svg.encode("utf-8")).decode("utf-8")
    html = r'<img src="data:image/svg+xml;base64,%s"  height="50" />' % b64
    st.write(html, unsafe_allow_html=True)
    for _ in range(margin_bottom):
        st.write("\n")


def check_api_credentials_set() -> None:
    """Check if API credential is set in session state."""
    st = _import_st()
    api_key = st.session_state.get("api_key")
    if api_key:
        if api_key.startswith("land_sk_"):
            return  # non-empty v2 api key
        elif st.session_state.get("api_secret"):
            return  # non-empty v1 api key and secret

    st.error("Please open the sidebar and enter your API credential first.")
    st.stop()


def check_endpoint_id_set() -> None:
    """Check if endpoint ID is set in session state."""
    st = _import_st()
    if st.session_state.get("endpoint_id"):
        return
    st.error("Please open the sidebar and enter your CloudInference endpoint ID first.")
    st.stop()


def get_api_credential_or_use_default() -> Tuple[Optional[str], Optional[str]]:
    """Get API credential (api key and api secret) from the session state.
    If the API credential is not set in the session state, use the default API credential from environment variables.
    The output could be either a v1 API key and secret, or a v2 API key depends on what is set in the session state.
    """
    st = _import_st()
    key, secret = (st.session_state["api_key"], st.session_state["api_secret"])
    if not key:
        return get_default_api_key(), secret
    return key, secret


def get_default_api_key() -> Optional[str]:
    """Get the default (free trial) API key."""
    default_key = os.environ.get(_DEFAULT_API_KEY_ENV_VAR)
    if not default_key:
        _LOGGER.warning(
            "The default API key is not set in the application. Please set your API key in the environment variable. E.g. export LANDINGAI_API_KEY=......"
        )
        return None
    return default_key


def render_api_config_form(
    render_endpoint_id: bool = False,
    default_key: str = "",
    default_secret: str = "",
    default_endpoint_id: str = "",
) -> None:
    """Show API credential and endpoint ID (optionally) configuration form.
    The form provides a way for users to set the API credential and endpoint ID in the session state.
    The values saved in the session state can be accessed by `get_api_credential_or_use_default()`.
    This is the source of truth for the API credential and endpoint ID in the App.
    """

    st = _import_st()
    # Set a default value in session state for the first time
    if "api_key" not in st.session_state:
        st.session_state["api_key"] = default_key
    if "api_secret" not in st.session_state:
        st.session_state["api_secret"] = default_secret
    if "endpoint_id" not in st.session_state:
        st.session_state["endpoint_id"] = default_endpoint_id

    def _update_credential(
        api_key: str, api_secret: str, endpoint_id: Optional[str]
    ) -> None:
        st.session_state["api_key"] = api_key
        st.session_state["api_secret"] = api_secret
        if endpoint_id:
            st.session_state["endpoint_id"] = endpoint_id

    with st.form("api_credential_form"):
        api_key = st.text_input(
            "LandingLens API Key",
            key="lnd_api_key",
            value=st.session_state["api_key"],
            help="If left empty, the free trial API key is used. The default key is a free trial key with a rate limit, i.e. X times per day.",
            type="password",
        )
        api_secret = st.text_input(
            "LandingLens API Secret",
            key="lnd_api_secret",
            value=st.session_state["api_secret"],
            help="Leave it empty if you use a v2 API key. Only the legacy API key (i.e. v1) requires API secret. The v2 API key always starts with 'land_sk_'.",
        )
        endpoint_id = None
        if render_endpoint_id:
            endpoint_id = st.text_input(
                "Cloud Deployment endpoint ID",
                key="lnd_endpoint_id",
                value=st.session_state.get("endpoint_id", ""),
            )

        submitted = st.form_submit_button("Save")
        if submitted:
            _update_credential(api_key, api_secret, endpoint_id)
            st.info("API configuration is saved successfully")

    key, _ = get_api_credential_or_use_default()
    if key == get_default_api_key():
        st.info("The default API key (free trial) is used")



import math
from collections import defaultdict
from itertools import groupby
from typing import Dict, List, Sequence, Tuple, Union, cast

from landingai.common import (
    ClassificationPrediction,
    ObjectDetectionPrediction,
    SegmentationPrediction,
)


def class_map(predictions: Sequence[ClassificationPrediction]) -> Dict[int, str]:
    """Return a map from the predicted class/label index to the class/label name."""
    return {pred.label_index: pred.label_name for pred in predictions}


def class_counts(
    predictions: Sequence[ClassificationPrediction],
) -> Dict[int, Tuple[int, str]]:
    """Compute the distribution of the occurrences of each class.

    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the number of occurrence, class/label name) as the value.
        ```
        Example:
            {
                1: (10, "cat"),
                2: (31, "dog"),
            }
        ```
    """
    counts: Dict[int, List[Union[int, str]]] = defaultdict(lambda: [0, ""])
    for pred in predictions:
        counts[pred.label_index][0] = cast(int, counts[pred.label_index][0]) + 1
        counts[pred.label_index][1] = pred.label_name
    return {k: (cast(int, v[0]), cast(str, v[1])) for k, v in counts.items()}


def class_pixel_coverage(
    predictions: Sequence[ClassificationPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    """Compute the pixel coverage of each class.

    Supported prediction types are:
    - SegmentationPrediction
    - ObjectDetectionPrediction

    It supports two ways to compute the coverage:
    - "absolute"
    - "relative"
    See the documentation of the coverage_type for more details.

    Parameters
    ----------
    predictions: a list of predictions. It could come from one or multiple images.
    coverage_type: "absolute" or "relative".
            - Absolute: The number of pixels of each predicted class.
            - Relative: The percentage of pixels that are predicted as the class
               over the sum total number of pixels of every mask. The only project type that supports "relative" is "SegmentationPrediction".


    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the coverage, class/label name) as the value.
        ```
        Example (coverage_type="absolute"):
            {
                0: (23512, "blue"),
                1: (1230, "green"),
                2: (0, "pink"),
            }
        ```
    """
    assert isinstance(
        predictions[0], SegmentationPrediction
    ), "Only support SegmentationPrediction for now."
    predictions = cast(List[SegmentationPrediction], predictions)
    return segmentation_class_pixel_coverage(predictions, coverage_type)


def od_class_pixel_coverage(
    predictions: Sequence[ObjectDetectionPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    raise NotImplementedError()


def segmentation_class_pixel_coverage(
    predictions: Sequence[SegmentationPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    """Compute the pixel coverage of each class.
    The coverage is defined as the percentage of pixels that are predicted as the class
    over the sum total number of pixels of every mask.
    If the predictions are from multiple images, the coverage is the average coverage across all images.

    Parameters
    ----------
    predictions: A list of segmentation predictions. It could come from one or multiple images.

    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the coverage percentage, class/label name) as the value.
        Note: The sum of the coverage percentage over all classes is not guaranteed
        to be 1.
        ```
        Example (coverage_type="relative"):
            {
                0: (0.15, "blue"),
                1: (0.31, "green"),
                2: (0.07, "pink"),
            }
        ```
    """
    pixel_coverages = []
    label_map = {}
    for pred in predictions:
        coverage = cast(float, pred.num_predicted_pixels)
        if coverage_type == "relative":
            coverage /= math.prod(pred.mask_shape)
        pixel_coverages.append((pred.label_index, coverage))
        label_map[pred.label_index] = pred.label_name

    sorted(pixel_coverages, key=lambda x: x[0])
    coverage_by_label: Dict[int, Tuple[float, str]] = {}
    for label_index, group in groupby(pixel_coverages, key=lambda x: x[0]):
        cov_vals = [item[1] for item in list(group)]
        avg_coverage = sum(cov_vals) / len(cov_vals) if len(cov_vals) > 0 else 0
        coverage_by_label[label_index] = (avg_coverage, label_map[label_index])
    return coverage_by_label



"""The vision pipeline abstraction helps chain image processing operations as sequence of steps. Each step consumes and produces a `FrameSet` which typically contains a source image and derivative metadata and images.
"""

import logging
import threading
import time
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Union, cast

import cv2
import numpy as np
from PIL import Image
from pydantic import BaseModel, PrivateAttr

from landingai.common import Prediction, ClassificationPrediction, in_notebook
from landingai.predict import Predictor
from landingai.visualize import overlay_predictions
from landingai.postprocess import class_counts


_LOGGER = logging.getLogger(__name__)


class Frame(BaseModel):
    """A Frame stores a main image, its metadata and potentially other derived images. This class will be mostly used internally by the FrameSet clase. A pipeline will `FrameSet` since it is more general and a can also keep new `Frames` extracted from existing ones"""

    image: Image.Image
    """Main image generated typically at the beginning of a pipeline"""

    other_images: Dict[str, Image.Image] = {}
    """Other derivative images associated with this the main image. For example: `FrameSet.overlay_predictions` will store the resulting image on `Frame.other_images["overlay"]"""

    predictions: List[ClassificationPrediction] = []
    """List of predictions for the main image"""

    metadata: Dict[str, Any] = {}
    """An optional collection of metadata"""

    def run_predict(self, predictor: Predictor) -> "Frame":
        """Run a cloud inference model
        Parameters
        ----------
        predictor: the model to be invoked.
        """
        self.predictions = predictor.predict(np.asarray(self.image))  # type: ignore
        return self

    def to_numpy_array(self) -> np.ndarray:
        """Return a numpy array using GRB color encoding (used by OpenCV)"""
        return np.asarray(self.image)

    class Config:
        arbitrary_types_allowed = True


class FrameSet(BaseModel):
    """A FrameSet is a collection of frames (in order). Typically a FrameSet will include a single image but there are circumstances where other images will be extracted from the initial one. For example: we may want to identify vehicles on an initial image and then extract sub-images for each of the vehicles."""

    frames: List[Frame] = []  # Start with empty frame set

    # @classmethod
    # def from_frameset_list(cls, list: List["FrameSet"] = None) -> "FrameSet":
    #     return cls(frames=list)

    @classmethod
    def from_image(
        cls, file: str, metadata: Optional[Dict[str, Any]] = {}
    ) -> "FrameSet":
        """Creates a FrameSet from an image file

        Parameters
        ----------
        file : Path to file

        Returns
        -------
        FrameSet : New FrameSet containing a single image
        """

        im = Image.open(file)
        return cls(frames=[Frame(image=im, metadata=metadata)])

    @classmethod
    def from_array(cls, array: np.ndarray, is_BGR: bool = True) -> "FrameSet":
        """Creates a FrameSet from a image encode as ndarray

        Parameters
        ----------
        array : np.ndarray
            Image
        is_BGR : bool, optional
            Assume OpenCV's BGR color space? Defaults to True

        Returns
        -------
        FrameSet
        """
        # img = cv2.cvtColor(np.asarray(self.frames[0].other_images[image_src]), cv2.COLOR_BGR2RGB)
        if is_BGR:
            array = cv2.cvtColor(array, cv2.COLOR_BGR2RGB)
        im = Image.fromarray(array)
        return cls(frames=[Frame(image=im)])

    # TODO: Is it worth to emulate a full container? - https://docs.python.org/3/reference/datamodel.html#emulating-container-types
    def __getitem__(self, key: int) -> Frame:
        return self.frames[key]

    def _repr_pretty_(self, pp, cycle) -> str:  # type: ignore
        # Enable a pretty output on Jupiter notebooks `Display()` function
        return str(
            pp.text(
                self.json(
                    # exclude={"frames": {"__all__": {"image", "other_images"}}},
                    indent=2
                )
            )
        )

    def is_empty(self) -> bool:
        """Check if the FrameSet is empty
        Returns
        -------
        bool
            True if the are no Frames on the FrameSet
        """
        return not self.frames  # True if the list is empty

    def run_predict(self, predictor: Predictor) -> "FrameSet":
        """Run a cloud inference model
        Parameters
        ----------
        predictor: the model to be invoked.
        """

        for frame in self.frames:
            frame.predictions = predictor.predict(np.asarray(frame.image))  # type: ignore
        return self

    def overlay_predictions(
        self, options: Optional[Dict[str, Any]] = None
    ) -> "FrameSet":  # TODO: Optional where to store
        for frame in self.frames:
            frame.other_images["overlay"] = overlay_predictions(
                cast(List[Prediction], frame.predictions), frame.image, options
            )
        return self

    def resize(
        self, width: Union[int, None] = None, height: Union[int, None] = None
    ) -> "FrameSet":  # TODO: Optional where to store
        """Returns a resized copy of this image. If width or height is missing the resize will preserve the aspect ratio
        Parameters
        ----------
        width: The requested width in pixels.
        height: The requested width in pixels.
        """
        if width is None and height is None:  # No resize needed
            return self
        for frame in self.frames:
            # Compute the final dimensions on the first image
            if width is None:
                width = int(height * float(frame.image.size[0] / frame.image.size[1]))  # type: ignore
            if height is None:
                height = int(width * float(frame.image.size[1] / frame.image.size[0]))
            frame.image = frame.image.resize((width, height))
        return self

    def downsize(
        self, width: Union[int, None] = None, height: Union[int, None] = None
    ) -> "FrameSet":  # TODO: Optional where to store
        """Resize only if the image is larger than the expected dimensions,
        Parameters
        ----------
        width: The requested width in pixels.
        height: The requested width in pixels.
        """
        if width is None and height is None:  # No resize needed
            return self
        for frame in self.frames:
            # Compute the final dimensions on the first image
            if width is None:
                width = int(height * float(frame.image.size[0] / frame.image.size[1]))  # type: ignore
            if height is None:
                height = int(width * float(frame.image.size[1] / frame.image.size[0]))
            if frame.image.size[0] > width or frame.image.size[1] > height:
                frame.image = frame.image.resize((width, height))
        return self

    def save_image(self, filename_prefix: str, image_src: str = "") -> "FrameSet":
        """Save all the images on the FrameSet to disk (as PNG)

        Parameters
        ----------
        filename_prefix : path and name prefix for the image file
        image_src : if empty the source image will be displayed. Otherwise the image will be selected from `other_images`
        """
        timestamp = datetime.now().strftime(
            "%Y%m%d-%H%M%S"
        )  # TODO saving faster than 1 sec will cause image overwrite
        c = 0
        for frame in self.frames:
            img = frame.image if image_src == "" else frame.other_images[image_src]
            img.save(f"{filename_prefix}_{timestamp}_{image_src}_{c}.png", format="PNG")
            c += 1
        return self

    def show_image(
        self, image_src: str = "", clear_nb_cell: bool = False
    ) -> "FrameSet":
        """Open an a window and display all the images.
        Parameters
        ----------
        image_src: if empty the source image will be displayed. Otherwise the image will be selected from `other_images`
        """
        # TODO: Should show be a end leaf?
        # Check if we are on a notebook context
        if in_notebook():
            from IPython import display

            for frame in self.frames:
                if clear_nb_cell:
                    display.clear_output(wait=True)
                if image_src == "":
                    display.display(frame.image)
                else:
                    display.display(frame.other_images[image_src])
        else:
            # Use PIL's implementation
            for frame in self.frames:
                if image_src == "":
                    frame.image.show()
                else:
                    frame.other_images[image_src].show()

        # # TODO: Implement image stacking when we have multiple frames (https://answers.opencv.org/question/175912/how-to-display-multiple-images-in-one-window/)
        # """Open an OpenCV window and display all the images. This call will stop the execution until a key is pressed.
        # Parameters
        # ----------
        # image_src: if empty the source image will be displayed. Otherwise the image will be selected from `other_images`
        # """
        # # OpenCV is full of issues when it comes to displaying windows (see https://stackoverflow.com/questions/6116564/destroywindow-does-not-close-window-on-mac-using-python-and-opencv)
        # cv2.namedWindow("image")
        # cv2.startWindowThread()
        # if image_src == "":
        #     img = cv2.cvtColor(np.asarray(self.frames[0].image), cv2.COLOR_BGR2RGB)
        # else:
        #     img = cv2.cvtColor(np.asarray(self.frames[0].other_images[image_src]), cv2.COLOR_BGR2RGB)
        # cv2.imshow("Landing AI - Press any key to exit", img)
        # cv2.waitKey(0) # close window when a key press is detected
        # cv2.waitKey(1)
        # cv2.destroyWindow('image')
        # for i in range (1,5):
        #     cv2.waitKey(1)

        return self

    def extend(self, frs: "FrameSet") -> "FrameSet":
        """Add a all the Frames from `frs` into this FrameSet

        Parameters
        ----------
        frs : FrameSet
            Framerset to be added at the end of the current one

        Returns
        -------
        FrameSet
        """
        self.frames.extend(frs.frames)
        return self

    def get_class_counts(self, add_id_to_classname: bool = False) -> Dict[str, int]:
        """This method returns the number of occurrences of each detected class in the FrameSet.

        Parameters
        ----------
        add_id_to_classname : bool, optional
            By default, detections with the same class names and different defect id will be counted as the same. Set to True if you want to count them separately

        Returns
        -------
        Dict[str, int]
            A dictionary with the counts
            ```
            Example:
                {
                    "cat": 10,
                    "dog": 3
                }
            ```
        """
        counts = {}
        for i in range(len(self.frames)):
            # Here is a sample return from class_counts: {1: (3, 'Heart'), 3: (3, 'Club'), 4: (3, 'Spade'), 2: (3, 'Diamond')}
            for k, v in class_counts(self.frames[i].predictions).items():
                if add_id_to_classname:  # This is useful if class names are not unique
                    class_name = f"{v[1]}_{k}"
                else:
                    class_name = v[1]
                if class_name not in counts:
                    counts[class_name] = v[0]
                else:
                    counts[class_name] += v[0]
        return counts

    def apply(self, function: Callable[[Frame], Frame] = lambda f: f) -> "FrameSet":
        """Apply a function to all frames

        Parameters
        ----------
        function: lambda function that takes individual frames and returned an updated frame
        """
        for i in range(len(self.frames)):
            self.frames[i] = function(self.frames[i])
        return self

    def filter(self, function: Callable[[Frame], bool] = lambda f: True) -> "FrameSet":
        """Evaluate a function on every frame and keep or remove

        Parameters
        ----------
        function : lambda function that gets invoked on every Frame. If it returns False, the Frame will be deleted
        """
        for i in reversed(
            range(0, len(self.frames))
        ):  # Traverse in reverse so we can delete
            if not function(self.frames[i]):
                self.frames.pop(i)
        return self

    class Config:
        # Add some encoders to prevent large structures from being printed
        json_encoders = {
            np.ndarray: lambda a: f"<np.ndarray: {a.shape}>",
            Image.Image: lambda i: f"<Image.Image: {i.size}>",
        }


# openCV's default VideoCapture cannot drop frames so if the CPU is overloaded the stream will tart to lag behind realtime.
# This class creates a treaded capture implementation that can stay up to date wit the stream and decodes frames only on demand
class NetworkedCamera(BaseModel):
    """The NetworkCamera class can connect to RTSP and other live video sources in order to grab frames. The main concern is to be able to consume frames at the source speed and drop them as needed to ensure the application allday gets the lastes frame"""

    stream_url: str
    motion_detection_threshold: int
    capture_interval: Union[float, None] = None
    previous_frame: Union[Frame, None] = None
    _last_capture_time: datetime = PrivateAttr()
    _cap: Any = PrivateAttr()  # cv2.VideoCapture
    _t: Any = PrivateAttr()  # threading.Thread
    _t_lock: Any = PrivateAttr()  # threading.Lock
    _t_running: bool = PrivateAttr()

    def __init__(
        self,
        stream_url: str,
        motion_detection_threshold: int = 0,
        capture_interval: Union[float, None] = None,
    ) -> None:
        """
        Parameters
        ----------
        stream_url : url to video source
        motion_detection_threshold : If set to zero then motion detections is disabled. Any other value (0-100) will make the camera drop all images that don't have significant changes
        capture_interval : If set to None, the NetworkedCamera will acquire images as fast as the source permits. Otherwise will grab a new frame every capture_interval seconds
        """
        cap = cv2.VideoCapture(stream_url)
        if not cap.isOpened():
            cap.release()
            raise Exception(f"Could not open stream ({stream_url})")
        cap.set(
            cv2.CAP_PROP_BUFFERSIZE, 2
        )  # Limit buffering to 2 frames in order to avoid backlog (i.e. lag)

        super().__init__(
            stream_url=stream_url,
            motion_detection_threshold=motion_detection_threshold,
            capture_interval=capture_interval,
        )
        self._last_capture_time = datetime.now()
        self._cap = cap
        self._t_lock = threading.Lock()
        self._t = threading.Thread(target=self._reader)
        self._t.daemon = True
        self._t_running = False
        self._t.start()

    def __del__(self) -> None:
        self._t_running = False
        self._t.join(timeout=10)
        self._cap.release()

    # grab frames as soon as they are available
    def _reader(self) -> None:
        self._t_running = True
        # inter_frame_interval = 1 / self._cap.get(cv2.CAP_PROP_FPS)  # Get the source's framerate (FPS = 1/X)
        inter_frame_interval = (
            1 / 30
        )  # Some sources miss report framerate so we use a conservative number
        while self._t_running:
            with self._t_lock:
                ret = self._cap.grab()  # non-blocking call
                if not ret:
                    raise Exception(f"Connection to camera broken ({self.stream_url})")
            time.sleep(inter_frame_interval)  # Limit acquisition speed

    # retrieve latest frame
    def get_latest_frame(self) -> "FrameSet":
        """Return the most up to date frame by dropping all by the latest frame. This function is blocking"""
        if self.capture_interval is not None:
            t = datetime.now()
            delta = (t - self._last_capture_time).total_seconds()
            if delta <= self.capture_interval:
                time.sleep(delta)
            self._last_capture_time = t
        with self._t_lock:
            ret, frame = self._cap.retrieve()  # non-blocking call
            if not ret:
                raise Exception(f"Connection to camera broken ({self.stream_url})")
        if self.motion_detection_threshold > 0:
            if self._detect_motion(frame):
                return FrameSet.from_array(frame)
            else:
                return FrameSet()  # Empty frame

        return FrameSet.from_array(frame)

    def _detect_motion(self, frame: np.ndarray) -> bool:  # TODO Needs test cases
        """ """
        # Prepare image; grayscale and blur
        prepared_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        prepared_frame = cv2.GaussianBlur(src=prepared_frame, ksize=(5, 5), sigmaX=0)

        if self.previous_frame is None:
            # Save the result for the next invocation
            self.previous_frame = prepared_frame
            return True  # First frame; there is no previous one yet

        # calculate difference and update previous frame TODO: don't assume the processed image is cached
        diff_frame = cv2.absdiff(src1=self.previous_frame, src2=prepared_frame)
        # Only take different areas that are different enough (>20 / 255)
        thresh_frame = cv2.threshold(
            src=diff_frame, thresh=20, maxval=255, type=cv2.THRESH_BINARY
        )[1]
        change_percentage = (
            100 * cv2.countNonZero(thresh_frame) / (frame.shape[0] * frame.shape[1])
        )
        # print(f"Image change {change_percentage:.2f}%")
        if change_percentage > self.motion_detection_threshold:
            self.previous_frame = prepared_frame
            return True
        return False

    # Make the class iterable. TODO Needs test cases
    def __iter__(self) -> Any:
        return self

    def __next__(self) -> "FrameSet":
        return self.get_latest_frame()

    class Config:
        arbitrary_types_allowed = True



"""Module for image transformations."""

from typing import List, Tuple

import cv2
import numpy as np
import PIL.Image


def crop_rotated_rectangle(
    image: PIL.Image.Image,
    rect: List[Tuple[int, int]],
    angle: float,
) -> Tuple[np.ndarray, List[Tuple[int, int]]]:
    """Crop the input image based on the rotated rectangle.
    The rectangle is calculated based on the rotated rectangle's corners.

    Parameters
    ----------
    img
       the input image to be cropped
    rect
        the unrotated rectangle (in parallel with the edges of the image)
    angle
        the angle of the rotated rectangle
    Returns
    -------
    Tuple[np.ndarray, List[Tuple[int, int]]]
        the cropped image and the coordinates of the rotated rectangle
    """
    # rot_matrix = cv2.getRotationMatrix2D(center=center, angle=angle, scale=1)
    # corners = cv2.transform(np.array(rect)[None], rot_matrix)
    # quad_box = corners[0].tolist()
    # return get_minarea_rect_crop(img, corners), quad_box
    [[left, top], [right, top], [right, bottom], [left, bottom]] = rect
    center = ((left + right) / 2, (top + bottom) / 2)
    img_np = np.asarray(image)
    shape = (img_np.shape[1], img_np.shape[0])

    matrix = cv2.getRotationMatrix2D(center=center, angle=angle, scale=1)
    img_np = cv2.warpAffine(src=img_np, M=matrix, dsize=shape)
    width, height = rect[1][0] - rect[0][0], rect[3][1] - rect[0][1]
    x, y = int(center[0] - width / 2), int(center[1] - height / 2)

    img_np = img_np[y : y + height, x : x + width]
    corners = cv2.transform(np.array(rect)[None], matrix)
    quad_box: List[Tuple[int, int]] = corners[0].tolist()

    return img_np, quad_box



"""A module that provides a set of abstractions and APIs for reading images from different sources."""

import glob
from collections.abc import Iterator
from pathlib import Path
from typing import Iterator as IteratorType
from typing import List, Union

from landingai.vision_pipeline import FrameSet


class ImageSourceBase(Iterator):
    """The base class for all image sources."""

    def __next__(self) -> FrameSet:
        raise NotImplementedError()


class ImageFolder(ImageSourceBase):
    """
    The `ImageFolder` class is an image source that reads images from a folder path.

    Example 1:
    ```
    folder = ImageFolder("/home/user/images")
    for image_batch in folder:
        print(image_batch[0].image.size)
    ```

    Example 2:
    ```
    # Read all jpg files in the folder (including nested files)
    folder = ImageFolder(glob_pattern="/home/user/images/**/*.jpg")
    for image_batch in folder:
        print(image_batch[0].image.size)
    ```
    """

    def __init__(
        self,
        source: Union[Path, str, List[str], None] = None,
        glob_pattern: Union[str, List[str], None] = None,
    ) -> None:
        """Constructor for ImageFolder.

        Parameters
        ----------
        source
            A list of file paths or the path to the folder path that contains the images.
            A folder path can be either an absolute or relative folder path, in `str` or `Path` type. E.g. "/home/user/images".
            If you provide a folder path, all the files directly within the folder will be read (including non-image files).
            Nested files and sub-directories will be ignored.
            Consider using `glob_pattern` if you need to:
              1. filter out unwanted files, e.g. your folder has both image and non-image files
              2. read nested image files, e.g. `/home/user/images/**/*.jpg`.
            The ordering of images is based on the file name alphabetically if source is a folder path.
            If source is a list of files, the order of the input files is preserved.
            Currently only local file paths are supported.
        glob_pattern
            One or more python glob pattern(s) to grab only wanted files in the folder. E.g. "/home/user/images/*.jpg".
            NOTE: If `glob_pattern` is provided, the `source` parameter is ignored.
            For more information about glob pattern, see https://docs.python.org/3/library/glob.html

        """
        self._source = source
        self._image_paths: List[str] = []

        if source is None and glob_pattern is None:
            raise ValueError("Either 'source' or 'glob_pattern' must be provided.")
        if glob_pattern is not None:
            if isinstance(glob_pattern, str):
                glob_pattern = [glob_pattern]
            for pattern in glob_pattern:
                self._image_paths.extend(list(glob.glob(pattern, recursive=True)))
            self._image_paths.sort()
        elif isinstance(source, list):
            self._image_paths = source
        else:
            assert isinstance(source, str) or isinstance(source, Path)
            p = Path(source)
            if not p.exists():
                raise ValueError(f"Path '{p}' does not exist.")
            self._image_paths = [str(x) for x in p.glob("*") if x.is_file()]
            self._image_paths.sort()

    def __iter__(self) -> IteratorType[FrameSet]:
        for img_path in self._image_paths:
            meta = {"image_path": img_path}
            yield FrameSet.from_image(img_path, metadata=meta)

    def __len__(self) -> int:
        return len(self._image_paths)

    def __repr__(self) -> str:
        return str(self._image_paths)

    @property
    def image_paths(self) -> List[str]:
        """Returns a list of image paths."""
        return self._image_paths
